% ----------------------------------------------------------------------
% Tcl utilities for manipulating attribute ranges
%
% Author: Drew van Camp (drew@@cs.toronto.edu)
%
% Copyright (c) 1996 The University of Toronto.
%
% See the file "copyright" for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL WARRANTIES.
% ----------------------------------------------------------------------

\documentclass{article}
\usepackage{moretext,amstex,alltt,varioref}
\usepackage{delve}

\newcommand{\range}{\textbf{Range}}

\begin{document}

\rcsInfo $Id: range.w,v 1.12.2.4 1996/05/28 17:06:16 drew Exp $

\title{Range: Manipulation of Delve Attribute Ranges}
\author{Drew van Camp (drew@@@@cs.toronto.edu)\\[1ex]
	Department of Computer Science\\
	University of Toronto\\
	6 Kings College Road\\
	Toronto ON, Canada, M5S 1A4}

\vfil
\maketitle
\vfil
\copyrightNotice{1996}
\vfil
\clearpage
\tableofcontents
\clearpage

\section{Introduction}
Each attribute in a \delve\ dataset has a specified \textbf{range},
consisting of a list of items, each of which is either a single
permitted value, or a set of permitted numerical values having the
form of an open, closed, or half-open interval of real numbers, or a
range of integers.  The bounds of a real interval can be ordinary
numbers, or one of ``\texttt{Inf}'', ``\texttt{-Inf}'', or
``\texttt{+Inf}'', with ``\texttt{Inf}'' representing infinity.  For
example, \texttt{[0,1)} represents the interval from 0 to 1, including
0, but not including 1, and \texttt{(0,Inf)} represents the set of
positive real numbers.  An integer range extending from \textbf{low}
to \textbf{high}, inclusive, is written as
\textbf{low\texttt{..}high}; \textbf{low} and \textbf{high} can be
infinite, as for real intervals.  For example, \texttt{1..Inf}
represents the positive integers.  For precise definitions of the
formats of integer and real numbers, see Section~\vref{sec:isint} and
Section~\vref{sec:isfloat}.

The utilities in this module can be used to parse ranges, obtain
their type or limits, see if the limits are opened or closed, or check
if a value falls in the range, As well, there are utilities for
checking if a value is an integer, a floating point number, or
infinity.

Before describing the individual pieces of software, we want to make a
comment at the top of the \tcl\ source file that tells how the file
was generated, and where the documentation is.

@o range.tcl @{
# This file, and its companion documentation file ``range.tex'', were
# automatically generated from the file ``range.w'' using
# ``nuweb''. Any changes to either file will be lost the next time
# ``nuweb'' is run.
#
# See the ``copyright'' file for information on usage and
# redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Id: range.w,v 1.12.2.4 1996/05/28 17:06:16 drew Exp $
@}

\section{The \texttt{range} command}
In order to minimize pollution of the name-space, there is only one
command for manipulating ranges. It does, however, have several
sub-commands (options) that enable it to perform many operations.

\subsection{Description}
\usage{range}{option range ?arg ...?}
This command provides several operations on a range.  \texttt{Range}
is an attribute range in any of the formats described in the
introduction.  If \texttt{range} does not correspond to any of those
formats, an error will be generated.  \texttt{Option} indicates what
to do with the range.  Valid valid options are:
\begin{options}
\item[range contains \textsl{range value}]
Return one if the \texttt{value} falls within the given
\texttt{range}.  If the \texttt{range} is an integer range, then
\texttt{value} must also be an integer. Similarly, for \texttt{real}
ranges, \texttt{value} must be a floating-point number.  If
\texttt{range} is enumerated, then \texttt{value} must exactly match
one of its elements.
\item[range finite \textsl{range}]
Return one if \texttt{range} is countable. A \texttt{Real} range is
never finite. An \texttt{integer} range is finite if and only if
neither of its limits are \texttt{Inf} or \texttt{-Inf}.  An
\texttt{enumerated} range is always finite.
\item[range limits \textsl{range}]
Return a list with two elements: the lower and upper limits for
\texttt{range}.  \texttt{Range} must be either \texttt{real} or
\texttt{integer}.  An error will be generated if \texttt{range} is
\texttt{enumerated}, since \texttt{enumerated} ranges do not have
limits.
\item[range list \textsl{range}]
If \texttt{range} is finite, return a list containing all the elements
that fall in the range.  If \texttt{range} is not finite, an error is
generated.
\item[range openlower \textsl{range}]
Return one if \texttt{range} has an open lower limit, i.e. if the
lower bound of the range is included in the range.
\item[range openupper \textsl{range}]
Return one if \texttt{range} has an open upper limit, i.e. if the
upper bound of the range is included in the range.
\item[range type \textsl{range}]
Return the type of \texttt{range}, one of: \texttt{real},
\texttt{integer}, or \texttt{enumerated}.
\end{options}

\subsection{Implementation}
The \texttt{range} command is implemented as a large switch statement
operating on the \texttt{option} argument. Each branch returns the
proper result, given \texttt{option}. The end of the procedure is
never reached.

@o range.tcl @{
proc range { option range args } {
    switch $option {
	@<``range contains'' branch@>
	@<``range finite'' branch@>
	@<``range limits'' branch@>
	@<``range list'' branch@>
	@<``range openlower'' branch@>
	@<``range openupper'' branch@>
	@<``range type'' branch@>
	@<default branch@>
    }
}
@}

\subsubsection{\texttt{range contains}}
This branch of the procedure handles the \texttt{contains} option,
that determines if a particular \texttt{value} falls within the
\texttt{range}. 

The first thing the branch does is check that exactly one optional
argument was passed, and saves it in the \texttt{value} variable.  It
then obtains the type of the \texttt{range} using \texttt{range type}.
It is assumed that \texttt{range type} will generate an error if the
range can't be parsed.  If the \texttt{range} is \texttt{enumerated},
it promptly searches through the \texttt{range} for \texttt{value},
and returns one if occurs, zero otherwise.

The other types are more complicated. They require first checking that
\texttt{value} is of the proper type, then that it falls within the
bounds of the \texttt{range}.  Special care has to be taken at the
boundaries, and if the bounds are \texttt{Inf} or \texttt{-Inf} (as
\texttt{Inf} causes all kinds of problems in the \texttt{expr}
command).  We only fall through to the end of the branch if
\texttt{value} falls within \texttt{range}.

@D ``range contains'' branch @{
contains {
    if { [llength $args] != 1 } {
	error "wrong \# args: should be \"range contains range value\""
    }
    set value [lindex $args 0 ] ;
    set type  [range type $range] ;

    if { [string compare $type enumerated] == 0 } {
	return [expr { [lsearch $range $value] >= 0 }] ;
    }

    if { [string compare $type integer] == 0 && ![isInt $value] } {
	return 0 ;
    } elseif { [string compare $type real] == 0 && ![isFloat $value] } {
	return 0 ;
    }

    set limits	[range limits $range] ;
    set lower	[lindex $limits 0] ;
    set upper	[lindex $limits 1] ;

    if { ( [isInf $lower] && [string compare [string index $lower 0] -] != 0 )
	 || ( ![range openlower $range] && $value == $lower)
	 || $value < $lower } {
	return 0 ;
    }

    if { ( [isInf $upper] && [string compare [string index $upper 0] -] == 0 )
	 || ( ![range openupper $range] && $value == $upper)
	 || $value > $upper } {
	return 0 ;
    }

    return 1 ;
}
@}


\subsubsection{\texttt{range finite}}
This branch of the procedure handles the \texttt{finite} option, that
checks whether or not the values that fall within the range are
countable.  

The first thing the branch does is check that no optional arguments
were passed.  Next it performs a switch on the type of the
\texttt{range}.  As with all of the other branches, it is assumed that
\texttt{range type} will generate an error if the range can't be
parsed.  If \texttt{range} is \texttt{enumerated}, it returns 1. If
\texttt{range} is \texttt{real} or \texttt{integer} the bounds are
checked. If the lower bound is $-\infty$, or the upper bound is $+\infty$,
we return 0.  If the lower bound is greater than or equal to the upper
bound, the range is countable (i.e. it has either 0 or one value), and
we return 1.  Now that all the bounds have been checked, we return 0
if the type is real, or 1 if it is integer.

@D ``range finite'' branch @{
finite	{
    if ![lempty $args] {
	error "wrong \# args: should be \"range finite range\""
    }
    set type [range type $range] ;

    if { [string compare $type enumerated] == 0 } {
	return 1 ;
    }

    set limits	[range limits $range] ;
    set lower	[lindex $limits 0] ;
    set upper	[lindex $limits 1] ;

    if { ( [isInf $lower] && [string compare [string index $lower 0] -] == 0 )
	 || ( [isInf $upper] && [string compare [string index $upper 0] -] != 0 ) } {
	return 0 ;
    }

    if { [isInf $lower] || [isInf $upper] || $lower >= $upper } {
	return 1 ;
    }

    expr { [string compare $type "integer"] == 0 } ;
}
@}

\subsubsection{\texttt{range limits}}
This branch of the procedure handles the \texttt{limits} option, that
returns the lower and upper limits of the range.  

The first thing the branch does is check that no optional arguments
were passed.  Next it performs a switch on the type of the
\texttt{range}.  As with all of the other branches, it is assumed that
\texttt{range type} will generate an error if the \texttt{range} can't
be parsed.  It then sets a regular expression that can be used to
extract the limits from the \texttt{range}. If the \texttt{range} is
\texttt{enumerated}, an error will be generated.  Finally, it calls
the \texttt{regexp} command, and returns the list containing the limits.

@D ``range limits'' branch @{
limits {
    if ![lempty $args] {
	error "wrong \# args: should be \"range limits range\""
    }
    switch [range type $range] {
	integer {
	    set regexp "\[ \t\]*\[-+\]?(\[0-9\])+\[ \t\]*\.\.\[ \t\]*\[-+\]?(\[0-9\]\[ \t\]*)+"
	    set regexp "\[ \t\]*(\[^ \t\]*)\[ \t\]*\\.\\.\[ \t\]*(\[^ \t\]*)\[ \t\]*" ;
        }
        real {
	    set regexp "\[\\\[\\(\]\[ \t\]*(\[^ \t\]*)\[ \t\]*,\[ \t\]*(\[^ \t\]*)\[ \t\]*\[]\\)\]" ;
        }
        enumerated {
	    error "range \"$range\" doesn't have limits"
	}
     }
     regexp $regexp $range all lower upper ;
     return [list $lower $upper] ;
}
@}

\subsubsection{\texttt{range list}}
This branch of the procedure handles the \texttt{list} option, that
returns a list containing all of the values that can occur in
\texttt{range}. Obviously, this is only valid for \texttt{finite}
ranges.  

The first thing the branch does is check that no optional arguments
were passed.  It then checks that the \texttt{range} is finite, and
returns an error if it isn't.  Next it performs a switch on the type
of the \texttt{range}.  As with all of the other branches, it is
assumed that \texttt{range type} will generate an error if the
\texttt{range} can't be parsed.  It then checks if the type of the
\texttt{range} is \texttt{enumerated} and, if it is, returns the
\texttt{range} itself.  For the other types of ranges
(\texttt{integer} and \texttt{real}, it checks if there are any values
in the range and returns an empty string if not.  If there is only one
value in the range, it returns that value.  All that \texttt{range}
can be after all those checks is an \texttt{integer} range with
multiple values.  We just iterate over all integer values in integer
\texttt{range} appending them to a list.  It is careful to (not)
include the range limits, depending on whether or not the
\texttt{range} is opened or closed.  

@D ``range list'' branch @{
list	{
    if ![lempty $args] {
	error "wrong \# args: should be \"range list range\""
    }
    if ![range finite $range] {
	error "can't list an infinite range: \"$range\"" ;
    }
    set type	[range type $range] ;

    if { [string compare $type enumerated] == 0 } {
	set result $range ;

    } else {
	set limits	[range limits $range] ;
	set lower	[lindex $limits 0] ;
	set upper	[lindex $limits 1] ;

	if { [isInf $lower] || [isInf $upper]
	     || ( $lower > $upper )
	     || ( $lower == $upper
		 && ( ![range openlower $range] || ![range openupper $range] ) ) } {
	    set result	{} ;
	} elseif { $lower == $upper } {
	    set result	$lower ;
	} else {
	    incr lower [expr { ![range openlower $range] }] ;
	    incr upper [range openupper $range] ;
	    set result {} ;
	    for { set idx $lower } { $idx < $upper } { incr idx } {
		lappend result $idx ;
	    }
	}
    }
    return $result ;
}
@}

\subsubsection{\texttt{range openlower}}
This branch of the procedure handles the \texttt{openlower}
option, that returns one if the lower limit of the \texttt{range} is
open (i.e. the lower limit is included in the \texttt{range}).  Since
the limits of enumerated ranges aren't defined, whether or not they
are open is also not defined.  

The first thing the branch does is check that no optional arguments
were passed.  It next checks if the \texttt{range} is
\texttt{enumerated}. If it is, it generates an error. Then it checks
if it is an \texttt{integer} range: they're always open. Finally it
sets up a regular expression for extracting the left brace of a
\texttt{real} range and extracts it.  If the brace is a \verb+[+ it
returns one, if it is a \verb+(+ it returns zero.


@D ``range openlower'' branch @{
openlower {
    if ![lempty $args] {
	error "wrong \# args: should be \"range openlower range\""
    }

    switch [range type $range] {
	"enumerated" {
	    error "enumerated ranges are neither opened or closed" ;
	}
	"integer" {
	    return 1 ;
	}
	"real" {
	    set regexp {([\[\(]).*} ;
	    regexp "^\[ \t]*$regexp" $range all brace ;
	    return [expr { [string compare $brace {[}] == 0 } ] ;
        }
    }
}
@}

\subsubsection{\texttt{range openupper}}
This branch of the procedure handles the \texttt{openupper}
option, that returns one if the upper limit of the \texttt{range} is
open (i.e. the upper limit is included in the \texttt{range}).  Since
the limits of enumerated ranges aren't defined, whether or not they
are open is also not defined.  

The first thing the branch does is check that no optional arguments
were passed.  It next checks if the \texttt{range} is
\texttt{enumerated}. If it is, it generates an error. Then it checks
if it is an \texttt{integer} range: they're always open. Finally it
sets up a regular expression for extracting the right brace of a
\texttt{real} range and extracts it.  If the brace is a \verb+}+ it
returns one, if it is a \verb+)+ it returns zero.

@D ``range openupper'' branch @{
openupper {
    if ![lempty $args] {
	error "wrong \# args: should be \"range openupper range\""
    }

    switch [range type $range] {
	"enumerated" {
	    error "enumerated ranges are neither opened or closed" ;
	}
	"integer" {
	    return 1 ;
	}
	"real" {
	    set regexp {.*([]\)])} ;
  	    regexp "$regexp\[ \t]*$" $range all brace ;
	    return [expr { [string compare $brace {]}] == 0 }] ;
        }
    }
}
@}

\subsubsection{\texttt{range type}}
This branch of the procedure handles the \texttt{type}
option, that returns the type of the range, one of: \texttt{real},
\texttt{integer}, or \texttt{enumerated}.

The first thing the branch does is check that no optional arguments
were passed.  It then sets up a regular expression to see if the range
is a \texttt{real} range (as defined in the introduction).  If it is,
it makes sure that the bounds are both floating point numbers, and
returns an error if they aren't.  Next it sets up a regular expression
to see if it is an \texttt{integer} range.  If it is, then it makes
sure that the limits are integers and generates an error if they
aren't.  If it is didn't match either regular expression, the range is
assumed to be enumerated.  This may be a bad assumption in some cases,
but there is no other way to do it.

@D ``range type'' branch @{
type {
    if ![lempty $args] {
	error "wrong \# args: should be \"range type range\""
    }

    set regexp "^\[ \t\]*\[\\\[\\(\](.*),(.*)\[]\\)\]\[ \t\]*\$" ;
    if [regexp $regexp $range all lower upper] {
	if { ![isFloat $lower] } {
	    error "expected floating-point limit in range \"$range\" but got \"$lower\""
	}
	if { ![isFloat $upper] } {
	    error "expected floating-point limit in range \"$range\" but got \"$upper\""
	}
	return real ;
    }

    set regexp "(.*)\\.\\.(.*)"
    if [regexp $regexp $range all lower upper] {
	if { ![isInt $lower] } {
	    error "expected integer limit in range \"$range\" but got \"$lower\""
	}
	if { ![isInt $upper] } {
	    error "expected integer limit in range \"$range\" but got \"$upper\""
	}
	return integer ;
    }

    return "enumerated" ;
}
@}

\subsubsection{\texttt{default branch}}
This branch of the procedure handles the default option, which is any
unknown one.  It simply generates an error listing all of the valid
options. 

@d default branch @{
default {
    error "bad option \"$option\": should be contains, finite, limits, list, openlower, openupper, or type"
}
@}

\section{The \texttt{isInt} command}		\label{sec:isint}
Frequently in \tcl\ you are given a string on which you want to
perform some arithmetic operation.  However, if the string isn't a
valid number, all heck can break loose.  The \texttt{isInt} command is
provided so that you can easily determine whether a string can be
interpreted as in integer \textit{before} you try to perform the
arithmetic operation.

\subsection{Description}
\usage{isInt}{value}
This procedure returns one if the given \texttt{value} is an
integer. An integer is defined as one or more digits in the range zero
through nine preceded by an optional \texttt{+} or \texttt{-} sign,
or infinity as described in section \vref{sec:isinf}.  Preceding and
trailing white space is ignored.

\subsection{Implementation} 
The procedure is straight-forward.  It does a simple regular
expression match against the \texttt{value} using a pattern that will
cover all cases mentioned \vpageref[above]{sec:isint}, plus checks
whether it is infinity.  Caution has to be taken to not include the
pattern for the surrounding white space in the regular expression
variable.  Because of the way \tcl\ handles tabs, the
\texttt{$\backslash$t} has to appear inside double quotes.

@o range.tcl @{
proc isInt { value } {
    set regexp {[-+]?([0-9])+}
    expr { [regexp "^\[ \t]*$regexp\[ \t]*$" $value] || [isInf $value] }
}
@}


\section{The \texttt{isFloat} command}	\label{sec:isfloat}
Frequently in \tcl\ you are given a string on which you want to
perform some arithmetic operation.  However, if the string isn't a
valid number, all heck can break loose.  The \texttt{isFloat} command
is provided so that you can easily determine whether a string can be
interpreted as in floating-point number \textit{before} you try to
perform the arithmetic operation.
\subsection{Description}
\usage{isFloat}{value}
This procedure returns one if the given \texttt{value} is a
floating point number. A floating point number is defined as:
\[
         [\ \mbox{\tt +}\ |\ \mbox{\tt -} \ ]\ 
         \{\ digit\ \}\ [\ {\tt .}\ \{\ digit\ \}\ ]\
         [\ (\ \mbox{\tt e}\ |\ \mbox{\tt E}\ )\ 
         [\ \mbox{\tt +}\ |\ \mbox{\tt -}\ ]\ digit\ \{\ digit\ \}\ ]
\]
with the restriction that at least one digit must appear, not 
counting digits after an ``{\tt e}'' or ``{\tt E}''.  As well, it may
be infinity  as described in section \vref{sec:isinf}.  Preceding and
trailing white space is ignored.

\subsection{Implementation}
The implementation is straight-forward, even if the regular expression
isn't.  The procedure builds up a really nasty expression that matches
the definition \vpageref[above]{sec:isfloat}.  It then checks if
\texttt{value} matches it \textsl{and} that at least one digit occurs.
It also checks whether the \texttt{value} is infinity.  Caution has to
be taken to not include the pattern for the surrounding white space in
the regular expression variable.  Because of the way \tcl\ handles
tabs, the \texttt{$\backslash$t} has to appear inside double quotes.

@o range.tcl @{
proc isFloat { value } {
    set regexp {[-+]?([0-9])*(\.([0-9])*)?([eE][-+]?[0-9]+)?}
    expr { ([regexp "^\[ \t]*$regexp\[ \t]*$" $value \
		 all integer allFraction fraction exponent]
	    && !([string compare $integer {}] == 0 
		 && [string compare $fraction {}] == 0))
	   || [isInf $value] }
}
@}

\section{The \texttt{isInf} command}		\label{sec:isinf}
In \tcl\, arithmetic expressions do not handle infinite numbers at
all.  This command extends \tcl\ by defining ``infinity'',
and allowing you to check whether a string value is infinite.
\subsection{Description}
\usage{isInf}{value}
This procedure returns one if the given \texttt{value} is
infinity. Infinity is defined as one of the strings \texttt{inf},
\texttt{+inf}, or \texttt{-inf}, where the case of the string does not
matter.  Preceding and trailing white space is ignored.

\subsection{Implementation}
The procedure is straight-forward.  It does a simple regular
expression match against the \texttt{value} using a pattern that will
cover all cases mentioned \vpageref[above]{sec:isinf}.  Caution has to
be taken to not include the pattern for the surrounding white space in
the regular expression variable.  Because of the way \tcl\ handles
tabs, the \texttt{$\backslash$t} has to appear inside double quotes.

@o range.tcl @{
proc isInf { value } {
    set regexp {[-+]?[Ii][Nn][Ff]} ;
    regexp "^\[ \t]*$regexp\[ \t]*$" $value ;
}
@}

\end{document}
